@@ -193,7 +193,7 @@ fp_device_get_property (GObject    *object,
       break;
 
     case PROP_FINGER_STATUS:
-      g_value_set_enum (value, priv->finger_status);
+      g_value_set_flags (value, priv->finger_status);
       break;
 
     case PROP_DRIVER:
@@ -545,7 +545,7 @@ fp_device_get_finger_status (FpDevice *device)
 {
   FpDevicePrivate *priv = fp_device_get_instance_private (device);
 
-  g_return_val_if_fail (FP_IS_DEVICE (device), FP_SCAN_TYPE_SWIPE);
+  g_return_val_if_fail (FP_IS_DEVICE (device), FP_FINGER_STATUS_NONE);
 
   return priv->finger_status;
 }
@@ -792,7 +792,7 @@ fp_device_close_finish (FpDevice     *device,
  * in. The driver may make use of this metadata, when e.g. storing the print on
  * device memory. It is undefined whether this print is filled in by the driver
  * and returned, or whether the driver will return a newly created print after
- * enrollment successed.
+ * enrollment succeeded.
  */
 void
 fp_device_enroll (FpDevice           *device,
@@ -1018,6 +1018,7 @@ fp_device_identify (FpDevice           *device,
   g_autoptr(GTask) task = NULL;
   FpDevicePrivate *priv = fp_device_get_instance_private (device);
   FpMatchData *data;
+  int i;
 
   task = g_task_new (device, cancellable, callback, user_data);
   if (g_task_return_error_if_cancelled (task))
@@ -1042,7 +1043,13 @@ fp_device_identify (FpDevice           *device,
   maybe_cancel_on_cancelled (device, cancellable);
 
   data = g_new0 (FpMatchData, 1);
-  data->gallery = g_ptr_array_ref (prints);
+  /* We cannot store the gallery directly, because the ptr array may not own
+   * a reference to each print. Also, the caller could in principle modify the
+   * GPtrArray afterwards.
+   */
+  data->gallery = g_ptr_array_new_full (prints->len, g_object_unref);
+  for (i = 0; i < prints->len; i++)
+    g_ptr_array_add (data->gallery, g_object_ref (g_ptr_array_index (prints, i)));
   data->match_cb = match_cb;
   data->match_data = match_data;
   data->match_destroy = match_destroy;
